@@ -43,6 +43,7 @@ The privsep daemon exits when the communication channel is closed,
 
 '''
 
+from concurrent import futures
 import enum
 import errno
 import io
@@ -64,10 +65,11 @@ import eventlet
 from oslo_config import cfg
 from oslo_log import log as logging
 from oslo_utils import importutils
+import six
 
+from oslo_privsep._i18n import _
 from oslo_privsep import capabilities
 from oslo_privsep import comm
-from oslo_privsep._i18n import _, _LE, _LI, _LW
 
 
 LOG = logging.getLogger(__name__)
@@ -181,8 +183,7 @@ class _ClientChannel(comm.ClientChannel):
             reply = self.send_recv((Message.PING.value,))
             success = reply[0] == Message.PONG
         except Exception as e:
-            LOG.exception(
-                _LE('Error while sending initial PING to privsep: %s'), e)
+            LOG.exception('Error while sending initial PING to privsep: %s', e)
             success = False
         if not success:
             msg = _('Privsep daemon failed to start')
@@ -211,8 +212,8 @@ class _ClientChannel(comm.ClientChannel):
             if LOG.isEnabledFor(record.levelno):
                 LOG.logger.handle(record)
         else:
-            LOG.warning(_LW('Ignoring unexpected OOB message from privileged '
-                            'process: %r'), msg)
+            LOG.warning('Ignoring unexpected OOB message from privileged '
+                        'process: %r', msg)
 
 
 def fdopen(fd, *args, **kwargs):
@@ -319,14 +320,14 @@ class RootwrapClientChannel(_ClientChannel):
             listen_sock.listen(1)
 
             cmd = context.helper_command(sockpath)
-            LOG.info(_LI('Running privsep helper: %s'), cmd)
+            LOG.info('Running privsep helper: %s', cmd)
             proc = subprocess.Popen(cmd, shell=False, stderr=_fd_logger())
             if proc.wait() != 0:
-                msg = (_LE('privsep helper command exited non-zero (%s)') %
+                msg = ('privsep helper command exited non-zero (%s)' %
                        proc.returncode)
                 LOG.critical(msg)
                 raise FailedToDropPrivileges(msg)
-            LOG.info(_LI('Spawned new privsep daemon via rootwrap'))
+            LOG.info('Spawned new privsep daemon via rootwrap')
 
             sock, _addr = listen_sock.accept()
             LOG.debug('Accepted privsep connection to %s', sockpath)
@@ -353,6 +354,9 @@ class Daemon(object):
         self.user = context.conf.user
         self.group = context.conf.group
         self.caps = set(context.conf.capabilities)
+        self.thread_pool = futures.ThreadPoolExecutor(
+            context.conf.thread_pool_size)
+        self.communication_error = None
 
     def run(self):
         """Run request loop. Sets up environment, then calls loop()"""
@@ -391,7 +395,7 @@ class Daemon(object):
         finally:
             capabilities.set_keepcaps(False)
 
-        LOG.info(_LI('privsep process running with uid/gid: %(uid)s/%(gid)s'),
+        LOG.info('privsep process running with uid/gid: %(uid)s/%(gid)s',
                  {'uid': os.getuid(), 'gid': os.getgid()})
 
         capabilities.drop_all_caps_except(self.caps, self.caps, [])
@@ -406,61 +410,103 @@ class Daemon(object):
 
         eff, prm, inh = capabilities.get_caps()
         LOG.info(
-            _LI('privsep process running with capabilities '
-                '(eff/prm/inh): %(eff)s/%(prm)s/%(inh)s'),
+            'privsep process running with capabilities '
+            '(eff/prm/inh): %(eff)s/%(prm)s/%(inh)s',
             {
                 'eff': fmt_caps(eff),
                 'prm': fmt_caps(prm),
                 'inh': fmt_caps(inh),
             })
 
-    def _process_cmd(self, cmd, *args):
+    def _process_cmd(self, msgid, cmd, *args):
+        """Executes the requested command in an execution thread.
+
+        This executes a call within a thread executor and returns the results
+        of the execution.
+
+        :param msgid: The message identifier.
+        :param cmd: The `Message` type indicating the command type.
+        :param args: The function, args, and kwargs if a Message.CALL type.
+        :return: A tuple of the return status, optional call output, and
+                 optional error information.
+        """
         if cmd == Message.PING:
             return (Message.PONG.value,)
 
-        elif cmd == Message.CALL:
+        try:
+            if cmd != Message.CALL:
+                raise ProtocolError(_('Unknown privsep cmd: %s') % cmd)
+
+            # Extract the callable and arguments
             name, f_args, f_kwargs = args
             func = importutils.import_class(name)
-
             if not self.context.is_entrypoint(func):
                 msg = _('Invalid privsep function: %s not exported') % name
                 raise NameError(msg)
 
             ret = func(*f_args, **f_kwargs)
             return (Message.RET.value, ret)
+        except Exception as e:
+            LOG.debug(
+                'privsep: Exception during request[%(msgid)s]: '
+                '%(err)s', {'msgid': msgid, 'err': e}, exc_info=True)
+            cls = e.__class__
+            cls_name = '%s.%s' % (cls.__module__, cls.__name__)
+            return (Message.ERR.value, cls_name, e.args)
 
-        raise ProtocolError(_('Unknown privsep cmd: %s') % cmd)
+    def _create_done_callback(self, msgid):
+        """Creates a future callback to receive command execution results.
+
+        :param msgid: The message identifier.
+        :return: A future reply callback.
+        """
+        channel = self.channel
+
+        def _call_back(result):
+            """Future execution callback.
+
+            :param result: The `future` execution and its results.
+            """
+            try:
+                reply = result.result()
+                LOG.debug('privsep: reply[%(msgid)s]: %(reply)s',
+                          {'msgid': msgid, 'reply': reply})
+                channel.send((msgid, reply))
+            except IOError:
+                self.communication_error = sys.exc_info()
+            except Exception as e:
+                LOG.debug(
+                    'privsep: Exception during request[%(msgid)s]: '
+                    '%(err)s', {'msgid': msgid, 'err': e}, exc_info=True)
+                cls = e.__class__
+                cls_name = '%s.%s' % (cls.__module__, cls.__name__)
+                reply = (Message.ERR.value, cls_name, e.args)
+                try:
+                    channel.send((msgid, reply))
+                except IOError:
+                    self.communication_error = sys.exc_info()
+
+        return _call_back
 
     def loop(self):
         """Main body of daemon request loop"""
-        LOG.info(_LI('privsep daemon running as pid %s'), os.getpid())
+        LOG.info('privsep daemon running as pid %s', os.getpid())
 
         # We *are* this context now - any calls through it should be
         # executed locally.
         self.context.set_client_mode(False)
 
         for msgid, msg in self.channel:
-            LOG.debug('privsep: request[%(msgid)s]: %(req)s',
-                      {'msgid': msgid, 'req': msg})
-            try:
-                reply = self._process_cmd(*msg)
-            except Exception as e:
-                LOG.debug(
-                    'privsep: Exception during request[%(msgid)s]: %(err)s',
-                    {'msgid': msgid, 'err': e}, exc_info=True)
-                cls = e.__class__
-                cls_name = '%s.%s' % (cls.__module__, cls.__name__)
-                reply = (Message.ERR.value, cls_name, e.args)
-
-            try:
-                LOG.debug('privsep: reply[%(msgid)s]: %(reply)s',
-                          {'msgid': msgid, 'reply': reply})
-                self.channel.send((msgid, reply))
-            except IOError as e:
-                if e.errno == errno.EPIPE:
+            error = self.communication_error
+            if error:
+                if error[1].errno == errno.EPIPE:
                     # Write stream closed, exit loop
                     break
-                raise
+                six.reraise(error)
+
+            # Submit the command for execution
+            future = self.thread_pool.submit(self._process_cmd, msgid, *msg)
+            future.add_done_callback(self._create_done_callback(msgid))
 
         LOG.debug('Socket closed, shutting down privsep daemon')
 
@@ -481,8 +527,8 @@ def helper_main():
     context = importutils.import_class(cfg.CONF.privsep_context)
     from oslo_privsep import priv_context   # Avoid circular import
     if not isinstance(context, priv_context.PrivContext):
-        LOG.fatal(_LE('--privsep_context must be the (python) name of a '
-                      'PrivContext object'))
+        LOG.fatal('--privsep_context must be the (python) name of a '
+                  'PrivContext object')
 
     sock = socket.socket(socket.AF_UNIX)
     sock.connect(cfg.CONF.privsep_sock_path)
@@ -503,7 +549,7 @@ def helper_main():
     # Channel is set up now, so move to in-band logging
     replace_logging(PrivsepLogHandler(channel))
 
-    LOG.info(_LI('privsep daemon starting'))
+    LOG.info('privsep daemon starting')
 
     try:
         Daemon(channel, context).run()
